package com.monkeycoder.monkeynes.hardware;

import com.monkeycoder.monkeynes.base.*;
import com.monkeycoder.monkeynes.gui.*;
import com.monkeycoder.monkeynes.hardware.mapper.*;

public class NesHardware implements Runnable
{
  // Internal Hardware ------------------------------------
  private NesCpu               cpu           = null;
  private NesPpu               ppu           = null;
  private NesRom               rom           = null;
  private MemoryMapperFactory  mapperFactory = null;
  private NesMainMemory        mainMem       = null;
  private NesVideoMemory       vidMem        = null;
  private MemoryMapper         mapper        = null;     
  private HighResolutionTimer  clock         = null;
  
  // External Hardware ------------------------------------
  private VideoScreen          vidScreen     = null;
  private NesController[]      controllers   = new NesController[2];  // note that ports are 0 and 1!
  
  private boolean              power         = false;
  private long                 cycleCount    = 0L;       // total cycle count
    
  public NesHardware()
  {
    mainMem       = new NesMainMemory();
    vidMem        = new NesVideoMemory();
    cpu           = new NesCpu( mainMem, vidMem );
    ppu           = new NesPpu( vidMem );
    mapperFactory = new MemoryMapperFactory( mainMem, vidMem );
    clock         = new HighResolutionTimer( 0L );
  }
  
  public void setClockSpeed( long inClockSpeed )
  {
    clock.setClockSpeed( inClockSpeed );
  }
  
  public void connectVideoScreen( VideoScreen inVidScreen )
  {
    vidScreen = inVidScreen;
  }  
  
  public void connectController( NesController inController, int inPort )
  {
    controllers[inPort] = inController;
  }   
  
  public void insertRom( NesRom inRom )
  {
    rom    = inRom;
    mapper = mapperFactory.getMapper( rom );
  }
  
  // Returns the ROM that was in use
  public NesRom removeRom()
  {
    NesRom tempRom = rom;
    mapper = null;
    rom    = null;
    return tempRom;
  }
  
  public boolean isPowerOn()
  {
    return power;
  }
  
  public void powerOn()
  {
    cpu.powerOn();
    ppu.powerOn();
    if( mapper != null )
      mapper.powerOn(); 
      
    power = true;   
  }
  
  public void powerOff()
  {
    cpu.powerOff();
    ppu.powerOff();
    if( mapper != null )
      mapper.powerOff();
      
    power = false;  
  }
  
  public void reset()
  {
    
  }
  
  // ----------------------------------------------------------------------------------
  // The main timer loop --------------------------------------------------------------
  // ----------------------------------------------------------------------------------
  public void run()
  {
    long startTime  = System.currentTimeMillis();
    long stopTime   = 0L;
    long i          = 0L;
    long cycleBatch = 0L;
    
    cycleCount      = 0L;
    clock.startClock();
    
    while( power )
    {
      cycleBatch = clock.getClocksPast();
      
      // Do all the cycles in the batch one at a time
      if( cycleBatch != 0 )
      {
        for( i = 0; i < cycleBatch; i++ )
        {
          cpu.doCycle();
          ppu.doCycle();
          mapper.doCycle();
        }

        // Catch a little break
        sleepPause( 1 );

        DebugLog.log( 5, "Cycles Complete: " + cycleBatch );
        cycleCount += cycleBatch;
      }
      else
      {
        // I can do other upkeep things here.
        
        // Catch a little break
        sleepPause( 1 );
      }  
    }    
    
    clock.stopClock();
    stopTime = System.currentTimeMillis();
    DebugLog.log( 3, "Total Cycles Complete: " + cycleCount );  
    DebugLog.log( 3, "Total Time:            " + (stopTime - startTime) + "ms" );  
    DebugLog.log( 3, "Effective Clock Speed: " + (long)((double)cycleCount / (double)(stopTime - startTime) * 1000.0) + "hz" );
  }  
  // ----------------------------------------------------------------------------------
  // END The main timer loop ----------------------------------------------------------
  // ----------------------------------------------------------------------------------
  
  private void sleepPause( int inMs )
  {
    try 
    { 
      Thread.sleep( inMs ); 
    } 
    catch( InterruptedException ie ) 
    {
      // Do nothing
    }
  }
  
  
  // Helper Access Functions ------------------------------------
  public NesCpu getCpu()
  {
    return cpu;
  }
  
  public NesPpu getPpu()
  {
    return ppu;
  }
  
  public VideoScreen getVideoScreen()
  {
    return vidScreen;
  }
  
  public NesRom getRom()
  {
    return rom;
  }
  
  public MemoryMapper getMemoryMapper()
  {
    return mapper;
  }
  
  public NesMainMemory getMainMemory()
  {
    return mainMem;
  }
  
  public NesVideoMemory getVideoMemory()
  {
    return vidMem;
  }
  
  // ------------------------------------------------------------
  
  public String toString()
  {
    StringBuffer sb = new StringBuffer();
    
    sb.append( "NesHardware -------------------------\n" );
    
    // Clock Speed
    sb.append( "  Clock Speed: " );
    sb.append( clock.getClockSpeed() );
    sb.append( "hz\n" );
    
    // Video Screen
    sb.append( "  VideoScreen: " );
    if( vidScreen != null )
      sb.append( "connected\n" );
    else
      sb.append( "not connected\n" );
    
    // Controllers
    sb.append( "  Controllers\n" );
    for( int i = 0; i < controllers.length; i++ )
    {
      sb.append( "    Port " );
      sb.append( i );
      sb.append( ": " );
      if( controllers[i] != null )
      {
        sb.append( controllers[i].getTypeName() );
        sb.append( " [" );
        sb.append( controllers[i].getType() );        
        sb.append( "]" );
      }  
      else
        sb.append( "not connected" );
      sb.append( "\n" );  
    }  
    
    // ROM
    sb.append( "  ROM\n" );
    if( rom != null )
    {
      sb.append( rom.toString() );
      sb.append( "\n" );
    }  
    else
      sb.append( "  not inserted\n" );
      
    // Mapper
    sb.append( "  Mapper: " );
    if( mapper != null )
      sb.append( mapper.toString() );
    else
      sb.append( "  information not available" );
      
    
    return sb.toString();
  }
  
}